from typing import Self

from .emit_intensity import EmitIntensity
from .phase import Phase


class Drive:
    phase: Phase
    intensity: EmitIntensity

    def __init__(self: Self, *, intensity: EmitIntensity, phase: Phase) -> None:
        self.phase = phase
        self.intensity = intensity

    NULL: "Drive" = None  # type: ignore[assignment]


Drive.NULL = Drive(phase=Phase.ZERO, intensity=EmitIntensity.MIN)



from pyautd3.ethercat.dc_sys_time import DcSysTime
from pyautd3.native_methods.autd3 import GPIOIn
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import TransitionModeWrap
from pyautd3.native_methods.utils import ConstantADT


class TransitionMode(metaclass=ConstantADT):
    SyncIdx: TransitionModeWrap = Base().transition_mode_sync_idx()

    @staticmethod
    def SysTime(sys_time: DcSysTime) -> TransitionModeWrap:  # noqa: N802
        return Base().transition_mode_sys_time(sys_time._inner)

    @staticmethod
    def GPIO(gpio: GPIOIn) -> TransitionModeWrap:  # noqa: N802
        return Base().transition_mode_gpio(gpio)

    Ext: TransitionModeWrap = Base().transition_mode_ext()
    Immediate: TransitionModeWrap = Base().transition_mode_immediate()

    def __new__(cls: type["TransitionMode"]) -> "TransitionMode":
        raise NotImplementedError

    NONE: TransitionModeWrap = Base().transition_mode_none()



from typing import Self

from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3capi import NativeMethods as LinkNop
from pyautd3.native_methods.autd3capi_driver import LinkPtr


class Nop(Link):
    def __init__(self: Self) -> None:
        super().__init__()

    def _resolve(self: Self) -> LinkPtr:
        return LinkNop().link_nop()



import ctypes
import enum

from pyautd3.native_methods.autd3 import DcSysTime

NUM_TRANS_IN_UNIT: int = 249


NUM_TRANS_IN_X: int = 18


NUM_TRANS_IN_Y: int = 14


TRANS_SPACING_MM: float = 10.16


DEVICE_HEIGHT_MM: float = 151.4


DEVICE_WIDTH_MM: float = 192.0


class DebugTypeTag(enum.IntEnum):
    None_ = 0
    BaseSignal = 1
    Thermo = 2
    ForceFan = 3
    Sync = 4
    ModSegment = 5
    ModIdx = 6
    StmSegment = 7
    StmIdx = 8
    IsStmMode = 9
    PwmOut = 10
    Direct = 11
    SysTimeEq = 12

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SamplingConfigTag(enum.IntEnum):
    Division = 0
    Frequency = 1
    Period = 2
    FrequencyNearest = 3
    PeriodNearest = 4

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class TransitionModeTag(enum.IntEnum):
    SyncIdx = 0
    SysTime = 1
    Gpio = 2
    Ext = 3
    Immediate = 4
    None_ = 0xFF

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class AUTDStatus(enum.IntEnum):
    AUTDTrue = 0
    AUTDFalse = 1
    AUTDErr = 2

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SleeperTag(enum.IntEnum):
    Std = 0
    Spin = 1
    Waitable = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SpinStrategyTag(enum.IntEnum):
    YieldThread = 0
    SpinLoopHint = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class DebugTypeValue(ctypes.Union):
    _fields_ = [("null", ctypes.c_uint64), ("sys_time", DcSysTime), ("idx", ctypes.c_uint16), ("direct", ctypes.c_bool)]


class SamplingConfigValue(ctypes.Union):
    _fields_ = [("division", ctypes.c_uint16), ("freq", ctypes.c_float), ("period_ns", ctypes.c_uint64)]


class TransitionModeValue(ctypes.Union):
    _fields_ = [("null", ctypes.c_uint64), ("sys_time", DcSysTime), ("gpio_in", ctypes.c_uint8)]


class ResultU16(ctypes.Structure):
    _fields_ = [("result", ctypes.c_uint16), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultU16) and self._fields_ == other._fields_  # pragma: no cover


class SleeperWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", ctypes.c_uint32), ("spin_strategy", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SleeperWrap) and self._fields_ == other._fields_  # pragma: no cover


class GainPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GainPtr) and self._fields_ == other._fields_  # pragma: no cover


class ControllerPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ControllerPtr) and self._fields_ == other._fields_  # pragma: no cover


class ResultStatus(ctypes.Structure):
    _fields_ = [("result", ctypes.c_uint8), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultStatus) and self._fields_ == other._fields_  # pragma: no cover


class Duration(ctypes.Structure):
    _fields_ = [("nanos", ctypes.c_uint64)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Duration) and self._fields_ == other._fields_  # pragma: no cover


class ResultDuration(ctypes.Structure):
    _fields_ = [("result", Duration), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultDuration) and self._fields_ == other._fields_  # pragma: no cover


class DatagramPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DatagramPtr) and self._fields_ == other._fields_  # pragma: no cover


class TransducerPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TransducerPtr) and self._fields_ == other._fields_  # pragma: no cover


class DevicePtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DevicePtr) and self._fields_ == other._fields_  # pragma: no cover


class ConstPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ConstPtr) and self._fields_ == other._fields_  # pragma: no cover


class SamplingConfigWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", SamplingConfigValue)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SamplingConfigWrap) and self._fields_ == other._fields_  # pragma: no cover


class LoopBehavior(ctypes.Structure):
    _fields_ = [("rep", ctypes.c_uint16)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LoopBehavior) and self._fields_ == other._fields_  # pragma: no cover


class SenderPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SenderPtr) and self._fields_ == other._fields_  # pragma: no cover


class GeometryPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GeometryPtr) and self._fields_ == other._fields_  # pragma: no cover


class ModulationPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ModulationPtr) and self._fields_ == other._fields_  # pragma: no cover


class ResultModulation(ctypes.Structure):
    _fields_ = [("result", ModulationPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultModulation) and self._fields_ == other._fields_  # pragma: no cover


class FociSTMPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FociSTMPtr) and self._fields_ == other._fields_  # pragma: no cover


class GainSTMPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GainSTMPtr) and self._fields_ == other._fields_  # pragma: no cover


class ResultF32(ctypes.Structure):
    _fields_ = [("result", ctypes.c_float), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultF32) and self._fields_ == other._fields_  # pragma: no cover


class TransitionModeWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", TransitionModeValue)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TransitionModeWrap) and self._fields_ == other._fields_  # pragma: no cover


class ResultSamplingConfig(ctypes.Structure):
    _fields_ = [("result", SamplingConfigWrap), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultSamplingConfig) and self._fields_ == other._fields_  # pragma: no cover


class DebugTypeWrap(ctypes.Structure):
    _fields_ = [("ty", ctypes.c_uint8), ("value", DebugTypeValue)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DebugTypeWrap) and self._fields_ == other._fields_  # pragma: no cover


class LinkPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LinkPtr) and self._fields_ == other._fields_  # pragma: no cover


class ResultGain(ctypes.Structure):
    _fields_ = [("result", GainPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultGain) and self._fields_ == other._fields_  # pragma: no cover


class OptionDuration(ctypes.Structure):
    _fields_ = [("has_value", ctypes.c_bool), ("value", Duration)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, OptionDuration) and self._fields_ == other._fields_  # pragma: no cover


class ResultLink(ctypes.Structure):
    _fields_ = [("result", LinkPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultLink) and self._fields_ == other._fields_  # pragma: no cover



from typing import Self

from pyautd3.driver.defined import Angle
from pyautd3.native_methods.autd3 import Phase as Phase_
from pyautd3.native_methods.autd3capi import NativeMethods as Base


class Phase:
    value: int

    def __init__(self: Self, phase: "int | Angle") -> None:
        match phase:
            case int():
                self.value = phase
            case Angle():
                self.value = int(Base().phase_from_rad(phase.radian()))
            case _:
                raise TypeError

    def radian(self: Self) -> float:
        return float(Base().phase_to_rad(self._inner()))

    def __eq__(self: Self, other: object) -> bool:
        return isinstance(other, Phase) and self.value == other.value

    def __str__(self: Self) -> str:
        return f"Phase({self.value})"

    def __repr__(self: Self) -> str:
        return f"Phase({self.value})"

    def _inner(self: Self) -> Phase_:
        return Phase_(self.value)

    ZERO: "Phase" = None  # type: ignore[assignment]
    PI: "Phase" = None  # type: ignore[assignment]


Phase.ZERO = Phase(0x00)
Phase.PI = Phase(0x80)



import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3capi_driver import ResultModulation, ResultStatus, SamplingConfigWrap


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi_modulation_audio_file{bin_ext}"))

        self.dll.AUTDModulationAudioFileTracingInit.argtypes = []
        self.dll.AUTDModulationAudioFileTracingInit.restype = None

        self.dll.AUTDModulationAudioFileTracingInitWithFile.argtypes = [ctypes.c_char_p]
        self.dll.AUTDModulationAudioFileTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDModulationAudioFileWav.argtypes = [ctypes.c_char_p]
        self.dll.AUTDModulationAudioFileWav.restype = ResultModulation

        self.dll.AUTDModulationAudioFileCsv.argtypes = [ctypes.c_char_p, SamplingConfigWrap, ctypes.c_uint8]
        self.dll.AUTDModulationAudioFileCsv.restype = ResultModulation

    def modulation_audio_file_tracing_init(self) -> None:
        return self.dll.AUTDModulationAudioFileTracingInit()

    def modulation_audio_file_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDModulationAudioFileTracingInitWithFile(path)

    def modulation_audio_file_wav(self, path: bytes) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileWav(path)

    def modulation_audio_file_csv(self, path: bytes, sampling_config: SamplingConfigWrap, delimiter: int) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileCsv(path, sampling_config, delimiter)



from pyautd3.driver.datagram.datagram import Datagram

from .clear import Clear
from .debug import DebugSettings, DebugType
from .force_fan import ForceFan
from .phase_corr import PhaseCorrection
from .pulse_width_encoder import PulseWidthEncoder
from .reads_fpga_state import ReadsFPGAState
from .segment import SwapSegment
from .silencer import FixedCompletionTime, FixedUpdateRate, Silencer
from .stm import FociSTM, GainSTM, GainSTMMode, GainSTMOption
from .synchronize import Synchronize
from .with_loop_behavior import WithLoopBehavior
from .with_segment import WithSegment

__all__ = [
    "Clear",
    "Datagram",
    "DebugSettings",
    "DebugType",
    "FixedCompletionTime",
    "FixedUpdateRate",
    "FociSTM",
    "ForceFan",
    "GainSTM",
    "GainSTMMode",
    "GainSTMOption",
    "PhaseCorrection",
    "PulseWidthEncoder",
    "ReadsFPGAState",
    "Silencer",
    "SwapSegment",
    "Synchronize",
    "WithLoopBehavior",
    "WithSegment",
]



import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3capi_driver import ResultLink, ResultStatus


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi_link_twincat{bin_ext}"))

        self.dll.AUTDLinkTwinCATTracingInit.argtypes = []
        self.dll.AUTDLinkTwinCATTracingInit.restype = None

        self.dll.AUTDLinkTwinCATTracingInitWithFile.argtypes = [ctypes.c_char_p]
        self.dll.AUTDLinkTwinCATTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDLinkTwinCAT.argtypes = []
        self.dll.AUTDLinkTwinCAT.restype = ResultLink

        self.dll.AUTDLinkRemoteTwinCAT.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
        self.dll.AUTDLinkRemoteTwinCAT.restype = ResultLink

    def link_twin_cat_tracing_init(self) -> None:
        return self.dll.AUTDLinkTwinCATTracingInit()

    def link_twin_cat_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDLinkTwinCATTracingInitWithFile(path)

    def link_twin_cat(self) -> ResultLink:
        return self.dll.AUTDLinkTwinCAT()

    def link_remote_twin_cat(self, server_ams_net_id: bytes, server_ip: bytes, client_ams_net_id: bytes) -> ResultLink:
        return self.dll.AUTDLinkRemoteTwinCAT(server_ams_net_id, server_ip, client_ams_net_id)



from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import GainCachePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr

G = TypeVar("G", bound=Gain)


class Cache(Gain, Generic[G]):
    _g: G
    _ptr: GainCachePtr | None

    def __init__(self: Self, g: G) -> None:
        super().__init__()
        self._g = g
        self._ptr = None

    def _gain_ptr(self: Self, geometry: Geometry) -> GainPtr:
        if self._ptr is None:
            self._ptr = Base().gain_cache(self._g._gain_ptr(geometry))
        return Base().gain_cache_clone(self._ptr)

    def __del__(self: Self) -> None:
        if self._ptr is not None:
            Base().gain_cache_free(self._ptr)



from .device import Device
from .geometry import Geometry
from .rotation import EulerAngles
from .transducer import Transducer

__all__ = [
    "Device",
    "EulerAngles",
    "Geometry",
    "Transducer",
]



from pyautd3.native_methods.autd3 import GainSTMMode

from .control_point import (
    ControlPoint,
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
)
from .foci import FociSTM
from .gain import GainSTM, GainSTMOption

__all__ = [
    "ControlPoint",
    "ControlPoints1",
    "ControlPoints2",
    "ControlPoints3",
    "ControlPoints4",
    "ControlPoints5",
    "ControlPoints6",
    "ControlPoints7",
    "ControlPoints8",
    "FociSTM",
    "GainSTM",
    "GainSTMMode",
    "GainSTMOption",
]



from typing import Self

from pyautd3.native_methods.autd3capi import FixedCompletionTime as FixedCompletionTime_
from pyautd3.utils import Duration


class FixedCompletionTime:
    intensity: Duration
    phase: Duration
    strict_mode: bool

    def __init__(self: Self, *, intensity: Duration | None = None, phase: Duration | None = None, strict_mode: bool = True) -> None:
        self.intensity = intensity or Duration.from_micros(250)
        self.phase = phase or Duration.from_micros(1000)
        self.strict_mode = strict_mode

    def _inner(self: Self) -> FixedCompletionTime_:
        return FixedCompletionTime_(self.intensity._inner, self.phase._inner, self.strict_mode)



from .bessel import Bessel, BesselOption
from .cache import Cache
from .custom import Custom
from .focus import Focus, FocusOption
from .group import Group
from .null import Null
from .plane import Plane, PlaneOption
from .uniform import Uniform

__all__ = [
    "Bessel",
    "BesselOption",
    "Cache",
    "Custom",
    "Focus",
    "FocusOption",
    "Group",
    "Null",
    "Plane",
    "PlaneOption",
    "Uniform",
]



import ctypes
from collections.abc import Iterable
from typing import Generic, Self, TypeVar

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.stm.control_point import (
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
    IControlPoints,
)
from pyautd3.driver.datagram.stm.stm_sampling_config import FreqNearest, PeriodNearest, _sampling_config
from pyautd3.driver.datagram.with_loop_behavior import DatagramL
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, FociSTMPtr, LoopBehavior, TransitionModeWrap
from pyautd3.utils import Duration

__all__ = []  # type: ignore[var-annotated]

C = TypeVar("C", bound=IControlPoints)


class FociSTM(
    DatagramS[FociSTMPtr],
    DatagramL[FociSTMPtr],
    Datagram,
    Generic[C],
):
    foci: list[C]
    config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest

    @classmethod
    def __private_new__(
        cls: type["FociSTM"],
        foci: (
            Iterable[ArrayLike]
            | Iterable[ControlPoints1]
            | Iterable[ControlPoints2]
            | Iterable[ControlPoints3]
            | Iterable[ControlPoints4]
            | Iterable[ControlPoints5]
            | Iterable[ControlPoints6]
            | Iterable[ControlPoints7]
            | Iterable[ControlPoints8]
        ),
        config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest,
    ) -> "FociSTM":
        ins = super().__new__(cls)
        ins.__private_init__(foci, config)
        return ins

    def __private_init__(
        self: "FociSTM",
        foci: (
            Iterable[ArrayLike]
            | Iterable[ControlPoints1]
            | Iterable[ControlPoints2]
            | Iterable[ControlPoints3]
            | Iterable[ControlPoints4]
            | Iterable[ControlPoints5]
            | Iterable[ControlPoints6]
            | Iterable[ControlPoints7]
            | Iterable[ControlPoints8]
        ),
        config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest,
    ) -> None:
        foci = list(foci)
        match foci[0]:
            case (
                ControlPoints1()
                | ControlPoints2()
                | ControlPoints3()
                | ControlPoints4()
                | ControlPoints5()
                | ControlPoints6()
                | ControlPoints7()
                | ControlPoints8()
            ):
                self.foci = foci
            case _:
                self.foci = [ControlPoints1(points=p) for p in foci]

        self.config = config

    def __init__(
        self: "FociSTM",
        foci: (
            Iterable[ArrayLike]
            | Iterable[ControlPoints1]
            | Iterable[ControlPoints2]
            | Iterable[ControlPoints3]
            | Iterable[ControlPoints4]
            | Iterable[ControlPoints5]
            | Iterable[ControlPoints6]
            | Iterable[ControlPoints7]
            | Iterable[ControlPoints8]
        ),
        config: SamplingConfig | Freq[float] | Duration,
    ) -> None:
        self.__private_init__(foci, config)

    def into_nearest(self: Self) -> "FociSTM":
        match self.config:
            case Freq() as freq:
                return FociSTM.__private_new__(self.foci, FreqNearest(freq))  # type: ignore[arg-type]
            case Duration() as period:
                return FociSTM.__private_new__(self.foci, PeriodNearest(period))  # type: ignore[arg-type]
            case _:
                raise TypeError

    def _raw_ptr(self: Self, _: Geometry) -> FociSTMPtr:
        n = self.foci[0]._value()
        foci = np.fromiter((np.void(p) for p in self.foci), dtype=np.dtype((np.void, 4 + n * 16)))  # type: ignore[type-var,call-overload]
        return Base().stm_foci(
            self.sampling_config()._inner,
            foci.ctypes.data_as(ctypes.c_void_p),  # type: ignore[arg-type]
            len(self.foci),
            n,
        )

    def _into_segment(self: Self, ptr: FociSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        return Base().stm_foci_into_datagram_with_segment(ptr, self.foci[0]._value(), segment, transition_mode or TransitionMode.NONE)

    def _into_loop_behavior(
        self: Self,
        ptr: FociSTMPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return Base().stm_foci_into_datagram_with_loop_behavior(
            ptr,
            self.foci[0]._value(),
            segment,
            transition_mode or TransitionMode.NONE,
            loop_behavior,
        )

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().stm_foci_into_datagram(self._raw_ptr(geometry), self.foci[0]._value())

    def sampling_config(self: Self) -> SamplingConfig:
        return _sampling_config(self.config, len(self.foci))



from ctypes import Array, c_float
from typing import Self

from pyautd3.gain.holo.backend import Backend
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import GSOption, GSPATOption, LMOption, NaiveOption
from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.structs import Point3


class NalgebraBackend(Backend):
    def __init__(self: Self) -> None:
        super().__init__(GainHolo().nalgebra_backend_sphere())

    def __del__(self: Self) -> None:
        if self._ptr.value is not None:
            GainHolo().delete_nalgebra_backend_sphere(self._ptr)
            self._ptr.value = None

    def _gs(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: GSOption) -> GainPtr:
        return GainHolo().gain_holo_gs_sphere(self._backend_ptr(), foci, amps, size, option)

    def _gspat(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: GSPATOption) -> GainPtr:
        return GainHolo().gain_holo_gspat_sphere(self._backend_ptr(), foci, amps, size, option)

    def _naive(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: NaiveOption) -> GainPtr:
        return GainHolo().gain_holo_naive_sphere(self._backend_ptr(), foci, amps, size, option)

    def _lm(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: LMOption) -> GainPtr:
        return GainHolo().gain_holo_lm_sphere(self._backend_ptr(), foci, amps, size, option)



from typing import Self

from pyautd3.driver.utils import _validate_nonzero_u16
from pyautd3.native_methods.autd3capi import FixedUpdateRate as FixedUpdateRate_


class FixedUpdateRate:
    intensity: int
    phase: int

    def __init__(self: Self, *, intensity: int, phase: int) -> None:
        self.intensity = _validate_nonzero_u16(intensity)
        self.phase = _validate_nonzero_u16(phase)

    def _inner(self: Self) -> FixedUpdateRate_:
        return FixedUpdateRate_(self.intensity, self.phase)



from pyautd3.autd_error import CantBeZeroError
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior
from pyautd3.native_methods.utils import ConstantADT


class LoopBehavior(metaclass=ConstantADT):
    Infinite: _LoopBehavior = Base().loop_behavior_infinite()

    @staticmethod
    def Finite(v: int) -> _LoopBehavior:  # noqa: N802
        if v < 1:
            raise CantBeZeroError(v)
        return Base().loop_behavior_finite(v)

    ONCE: _LoopBehavior = Finite(1)

    def __new__(cls: type["LoopBehavior"]) -> "LoopBehavior":
        raise NotImplementedError



from .cache import Cache
from .custom import Custom
from .fir import Fir
from .fourier import Fourier, FourierOption
from .radiation_pressure import RadiationPressure
from .sine import Sine, SineOption
from .square import Square, SquareOption
from .static import Static

__all__ = [
    "Cache",
    "Custom",
    "Fir",
    "Fourier",
    "FourierOption",
    "RadiationPressure",
    "Sine",
    "SineOption",
    "Square",
    "SquareOption",
    "Static",
]



import numpy as np
import pytest

from pyautd3 import AUTD3, Controller, EulerAngles, deg, rad
from pyautd3.driver.geometry.rotation import Angle
from pyautd3.link.audit import Audit

from .test_autd import create_controller


def test_angle():
    assert (np.pi / 2 * rad).radian() == np.pi / 2
    assert (90 * deg).radian() == np.pi / 2


def test_angle_ctr():
    with pytest.raises(NotImplementedError):
        _ = Angle()

    with pytest.raises(NotImplementedError):
        _ = Angle._UnitDegree()

    with pytest.raises(NotImplementedError):
        _ = Angle._UnitRad()

    with pytest.raises(NotImplementedError):
        _ = EulerAngles()


def test_with_rotation_xyz():
    def open_with_rotation(q: np.ndarray) -> Controller[Audit]:
        return Controller.open([AUTD3(pos=[0.0, 0.0, 0.0], rot=q)], Audit())

    with open_with_rotation(EulerAngles.XYZ(90 * deg, 0 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [0.0, -1.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 0 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 90 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(90 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [1.0, 0.0, 0.0])


def test_with_rotation_zyz():
    def open_with_rotation(q: np.ndarray) -> Controller[Audit]:
        return Controller.open([AUTD3(pos=[0.0, 0.0, 0.0], rot=q)], Audit())

    with open_with_rotation(EulerAngles.ZYZ(90 * deg, 0 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 0 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 90 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.ZYZ(90 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry()[0].x_direction(), [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry()[0].y_direction(), [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry()[0].axial_direction(), [0.0, 1.0, 0.0])


def test_geometry_num_devices():
    with create_controller() as autd:
        assert autd.num_devices() == 2


def test_geometry_center():
    with create_controller() as autd:
        center = autd.center()
        assert len(center) == 3
        assert center[0] == 86.625267028808594
        assert center[1] == 66.71319580078125
        assert center[2] == 0.0


def test_device_idx():
    with create_controller() as autd:
        assert autd.geometry()[0].idx() == 0
        assert autd.geometry()[1].idx() == 1


def test_device_sound_speed():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert dev.sound_speed == 340e3
            dev.sound_speed = 350e3
            assert dev.sound_speed == 350e3


def test_device_set_sound_speed_from_temp():
    with create_controller() as autd:
        for dev in autd.geometry():
            dev.set_sound_speed_from_temp(15)
            assert dev.sound_speed == 340.29525e3

        autd.set_sound_speed(350e3)
        for dev in autd.geometry():
            assert dev.sound_speed == 350e3

        autd.set_sound_speed_from_temp(15)
        for dev in autd.geometry():
            assert dev.sound_speed == 340.29525e3


def test_device_enable():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert dev.enable
            dev.enable = False
            assert not dev.enable


def test_device_num_transducers():
    with create_controller() as autd:
        assert autd.num_transducers() == 249 * autd.num_devices()
        for dev in autd.geometry():
            assert dev.num_transducers() == 249


def test_device_center():
    with create_controller() as autd:
        for dev in autd.geometry():
            center = dev.center()
            assert len(center) == 3
            assert center[0] == 86.625267028808594
            assert center[1] == 66.71319580078125
            assert center[2] == 0.0


def test_device_translate():
    with create_controller() as autd:
        for dev in autd.geometry():
            original_pos = [tr.position() for tr in dev]
            t = [1, 2, 3]
            dev.translate(t)
            for tr in dev:
                assert np.allclose(tr.position(), original_pos[tr.idx()] + t)


def test_device_rotate():
    with create_controller() as autd:
        for dev in autd.geometry():
            r = [0.70710678, 0.0, 0.0, 0.70710678]
            dev.rotate(r)
            assert np.allclose(dev.rotation(), r)


def test_device_affine():
    with create_controller() as autd:
        for dev in autd.geometry():
            original_pos = [tr.position() for tr in dev]
            t = [1, 2, 3]
            r = [0.70710678, 0.0, 0.0, 0.70710678]
            dev.affine(t, r)
            for tr in dev:
                op = original_pos[tr.idx()]
                expected = np.array([-op[1], op[0], op[2]]) + t
                assert np.allclose(tr.position(), expected, atol=1e-3)
            assert np.allclose(dev.rotation(), r)


def test_device_wavelength():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert dev.wavelength() == 340e3 / 40e3


def test_device_wavenum():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert dev.wavenumber() == 0.7391983270645142


def test_transducer_idx():
    with create_controller() as autd:
        for dev in autd.geometry():
            for i, tr in enumerate(dev):
                assert tr.idx() == i


def test_transducer_dev_idx():
    with create_controller() as autd:
        for dev in autd.geometry():
            for tr in dev:
                assert tr.dev_idx() == dev.idx()


def test_transducer_position():
    with create_controller() as autd:
        assert np.allclose(autd.geometry()[0][0].position(), [0.0, 0.0, 0.0])
        assert np.allclose(
            autd.geometry()[0][-1].position(),
            [(AUTD3.NUM_TRANS_X - 1) * AUTD3.TRANS_SPACING, (AUTD3.NUM_TRANS_Y - 1) * AUTD3.TRANS_SPACING, 0.0],
        )

        assert np.allclose(autd.geometry()[1][0].position(), [0.0, 0.0, 0.0])
        assert np.allclose(
            autd.geometry()[1][-1].position(),
            [(AUTD3.NUM_TRANS_X - 1) * AUTD3.TRANS_SPACING, (AUTD3.NUM_TRANS_Y - 1) * AUTD3.TRANS_SPACING, 0.0],
        )


def test_transducer_rotation():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.allclose(dev.rotation(), [1.0, 0.0, 0.0, 0.0])


def test_transducer_x_direction():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.allclose(dev.x_direction(), [1.0, 0.0, 0.0])


def test_transducer_y_direction():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.allclose(dev.y_direction(), [0.0, 1.0, 0.0])


def test_transducer_axial_direction():
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.allclose(dev.axial_direction(), [0.0, 0.0, 1.0])



from enum import Enum
from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.angle import Angle, rad
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SineOption as SineOption_
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

T = TypeVar("T", int, float)


class SineOption:
    intensity: int
    offset: int
    phase: Angle
    clamp: bool
    sampling_config: SamplingConfig

    def __init__(
        self: Self,
        *,
        intensity: int = 0xFF,
        offset: int = 0x80,
        phase: Angle | None = None,
        clamp: bool = False,
        sampling_config: SamplingConfig | None = None,
    ) -> None:
        self.intensity = _validate_u8(intensity)
        self.offset = _validate_u8(offset)
        self.phase = phase or 0.0 * rad
        self.clamp = clamp
        self.sampling_config = sampling_config or SamplingConfig.FREQ_4K

    def _inner(self: Self) -> SineOption_:
        return SineOption_(
            self.intensity,
            self.offset,
            self.phase._inner(),
            self.clamp,
            self.sampling_config.division,
        )


class SineMode(Enum):
    Exact = 0
    ExactFloat = 1
    Nearest = 2


class Sine(Modulation, Generic[T]):
    _mode: SineMode
    freq: Freq[T]
    option: SineOption

    def __init__(self: Self, freq: Freq[T], option: SineOption) -> None:
        super().__init__()
        match freq.hz():
            case int():
                self._mode = SineMode.Exact
            case _:
                self._mode = SineMode.ExactFloat
        self.freq = freq
        self.option = option

    def into_nearest(self: Self) -> "Sine":
        match self._mode:
            case SineMode.ExactFloat:
                new = Sine(self.freq, self.option)
                new._mode = SineMode.Nearest
                return new
            case _:
                raise TypeError

    def _modulation_ptr(self) -> ModulationPtr:
        match self._mode:
            case SineMode.Exact:
                return Base().modulation_sine_exact(self.freq.hz(), self.option._inner())  # type: ignore[arg-type]
            case SineMode.ExactFloat:
                return Base().modulation_sine_exact_float(self.freq.hz(), self.option._inner())
            case SineMode.Nearest:  # pragma: no cover
                return Base().modulation_sine_nearest(self.freq.hz(), self.option._inner())



from typing import Self

from pyautd3.driver.utils import _validate_nonzero_u16
from pyautd3.native_methods.autd3capi import FixedCompletionSteps as FixedCompletionSteps_


class FixedCompletionSteps:
    intensity: int
    phase: int
    strict_mode: bool

    def __init__(self: Self, *, intensity: int = 10, phase: int = 40, strict_mode: bool = True) -> None:
        self.intensity = _validate_nonzero_u16(intensity)
        self.phase = _validate_nonzero_u16(phase)
        self.strict_mode = strict_mode

    def _inner(self: Self) -> FixedCompletionSteps_:
        return FixedCompletionSteps_(self.intensity, self.phase, self.strict_mode)



from typing import Self

from pyautd3.native_methods.autd3capi import NativeMethods as Base


class FirmwareInfo:
    info: str

    def __init__(self: Self, info: str) -> None:
        self.info = info

    @staticmethod
    def latest_version() -> str:
        sb = bytes(bytearray(256))
        Base().firmware_latest(sb)
        return sb.decode("utf-8").rstrip(" \t\r\n\0")

    def __str__(self: Self) -> str:
        return self.info



import ctypes
import enum

from pyautd3.native_methods.structs import Point3


class ParallelMode(enum.IntEnum):
    Auto = 0
    On = 1
    Off = 2

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GPIOOut(enum.IntEnum):
    O0 = 0
    O1 = 1
    O2 = 2
    O3 = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GPIOIn(enum.IntEnum):
    I0 = 0
    I1 = 1
    I2 = 2
    I3 = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class Segment(enum.IntEnum):
    S0 = 0
    S1 = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GainSTMMode(enum.IntEnum):
    PhaseIntensityFull = 0
    PhaseFull = 1
    PhaseHalf = 2

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SilencerTarget(enum.IntEnum):
    Intensity = 0
    PulseWidth = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class EmitIntensity(ctypes.Structure):
    _fields_ = [("value", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmitIntensity) and self._fields_ == other._fields_  # pragma: no cover


class FixedCompletionSteps(ctypes.Structure):
    _fields_ = [("intensity", ctypes.c_uint16), ("phase", ctypes.c_uint16), ("strict_mode", ctypes.c_bool)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FixedCompletionSteps) and self._fields_ == other._fields_  # pragma: no cover


class FixedUpdateRate(ctypes.Structure):
    _fields_ = [("intensity", ctypes.c_uint16), ("phase", ctypes.c_uint16)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FixedUpdateRate) and self._fields_ == other._fields_  # pragma: no cover


class GainSTMOption(ctypes.Structure):
    _fields_ = [("mode", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GainSTMOption) and self._fields_ == other._fields_  # pragma: no cover


class FPGAState(ctypes.Structure):
    _fields_ = [("state", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FPGAState) and self._fields_ == other._fields_  # pragma: no cover


class DcSysTime(ctypes.Structure):
    _fields_ = [("dc_sys_time", ctypes.c_uint64)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DcSysTime) and self._fields_ == other._fields_  # pragma: no cover


class Phase(ctypes.Structure):
    _fields_ = [("value", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Phase) and self._fields_ == other._fields_  # pragma: no cover


class Angle(ctypes.Structure):
    _fields_ = [("radian", ctypes.c_float)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Angle) and self._fields_ == other._fields_  # pragma: no cover


class BesselOption(ctypes.Structure):
    _fields_ = [("intensity", EmitIntensity), ("phase_offset", Phase)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BesselOption) and self._fields_ == other._fields_  # pragma: no cover


class FocusOption(ctypes.Structure):
    _fields_ = [("intensity", EmitIntensity), ("phase_offset", Phase)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FocusOption) and self._fields_ == other._fields_  # pragma: no cover


class PlaneOption(ctypes.Structure):
    _fields_ = [("intensity", EmitIntensity), ("phase_offset", Phase)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, PlaneOption) and self._fields_ == other._fields_  # pragma: no cover


class ControlPoint(ctypes.Structure):
    _fields_ = [("point", Point3), ("phase_offset", Phase)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ControlPoint) and self._fields_ == other._fields_  # pragma: no cover


class Drive(ctypes.Structure):
    _fields_ = [("phase", Phase), ("intensity", EmitIntensity)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Drive) and self._fields_ == other._fields_  # pragma: no cover



from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.silencer.fixed_completion_steps import FixedCompletionSteps
from pyautd3.driver.datagram.silencer.fixed_completion_time import FixedCompletionTime
from pyautd3.driver.datagram.silencer.fixed_update_rate import FixedUpdateRate
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import SilencerTarget
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr

T = TypeVar("T", FixedCompletionSteps, FixedCompletionTime, FixedUpdateRate)


class Silencer(
    Datagram,
    Generic[T],
):
    config: T
    target: SilencerTarget

    def __init__(self: Self, config: T | None = None, target: SilencerTarget = SilencerTarget.Intensity) -> None:
        super().__init__()
        self.config = config or FixedCompletionSteps()  # type: ignore[assignment]
        self.target = target

    def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
        match self.config:
            case FixedCompletionSteps():
                return Base().datagram_silencer_from_completion_steps(self.config._inner(), self.target)
            case FixedCompletionTime():
                return Base().datagram_silencer_from_completion_time(self.config._inner(), self.target)
            case FixedUpdateRate():  # pragma: no cover
                return Base().datagram_silencer_from_update_rate(self.config._inner(), self.target)

    @staticmethod
    def disable() -> "Silencer[FixedCompletionSteps]":
        return Silencer(config=FixedCompletionSteps(intensity=1, phase=1), target=SilencerTarget.Intensity)



from typing import Self

from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3 import EmitIntensity as EmitIntensity_


class EmitIntensity:
    value: int

    def __init__(self: Self, intensity: int) -> None:
        self.value = _validate_u8(intensity)

    def __floordiv__(self: Self, other: int) -> "EmitIntensity":
        return EmitIntensity(self.value // other)

    def __eq__(self: Self, __value: object, /) -> bool:
        return isinstance(__value, EmitIntensity) and self.value == __value.value

    def __str__(self: Self) -> str:
        return f"EmitIntensity({self.value})"

    def __repr__(self: Self) -> str:
        return f"EmitIntensity({self.value})"

    def _inner(self: Self) -> EmitIntensity_:
        return EmitIntensity_(self.value)

    MIN: "EmitIntensity" = None  # type: ignore[assignment]
    MAX: "EmitIntensity" = None  # type: ignore[assignment]


EmitIntensity.MIN = EmitIntensity(0x00)
EmitIntensity.MAX = EmitIntensity(0xFF)



from typing import Self

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr


class Clear(Datagram):
    def __init__(self: Self) -> None:
        super().__init__()

    def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
        return Base().datagram_clear()



import platform

import numpy as np
import pytest

from pyautd3 import AUTD3, Clear, Controller, Device, ForceFan, Segment, tracing_init
from pyautd3.autd_error import AUTDError, InvalidDatagramTypeError
from pyautd3.controller.controller import SenderOption
from pyautd3.controller.sleeper import SpinSleeper, SpinStrategy, StdSleeper, WaitableSleeper
from pyautd3.driver.datagram import Synchronize
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.gain import Null, Uniform
from pyautd3.link.audit import Audit
from pyautd3.modulation import Sine, Static
from pyautd3.modulation.sine import SineOption
from pyautd3.native_methods.autd3capi import NativeMethods as Base


def create_controller() -> Controller[Audit]:
    return Controller.open(
        [AUTD3(pos=[0.0, 0.0, 0.0], rot=[1.0, 0.0, 0.0, 0.0]), AUTD3(pos=[0.0, 0.0, 0.0], rot=[1.0, 0.0, 0.0, 0.0])],
        Audit(),
    )


def test_sender_is_default():
    assert Base().sender_option_is_default(SenderOption()._inner())


def test_sleeper():
    _ = StdSleeper(timer_resolution=1)._inner()
    _ = SpinSleeper().with_spin_strategy(SpinStrategy.SpinLoopHint)._inner()
    if platform.system() == "Windows":
        _ = WaitableSleeper()._inner()
    else:
        with pytest.raises(RuntimeError):
            _ = WaitableSleeper()._inner()


def test_firmware_info():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert FirmwareInfo.latest_version() == "v10.0.1"

        for i, firm in enumerate(autd.firmware_version()):
            assert firm.info == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"
            assert str(firm) == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"


def test_close():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.link().is_open()

        autd.close()
        autd.close()

    with create_controller() as autd:
        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.close()
        assert str(e.value) == "broken"


def test_send_single():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)

        autd.send(Static())

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)

        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.send(Static())
        assert str(e.value) == "broken"
        autd.link().repair()


def test_send_tuple():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)

        autd.send((Static(), Uniform(intensity=EmitIntensity(0x80), phase=Phase(0))))
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0x80)
            assert np.all(phases == 0)

        with pytest.raises(InvalidDatagramTypeError):
            autd.send(0)  # type: ignore[arg-type]

        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.send((Static(), Uniform(intensity=EmitIntensity(0xFF), phase=Phase(0))))
        assert str(e.value) == "broken"
        autd.link().repair()


def test_group():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.group_send(
            key_map=lambda dev: dev.idx(),
            data_map={1: Null(), 0: (Sine(freq=150 * Hz, option=SineOption()), Uniform(intensity=EmitIntensity(0xFF), phase=Phase(0)))},
        )

        mod = autd.link().modulation_buffer(0, Segment.S0)
        assert len(mod) == 80
        intensities, phases = autd.link().drives_at(0, Segment.S0, 0)
        assert np.all(intensities == 0xFF)
        assert np.all(phases == 0)

        mod = autd.link().modulation_buffer(1, Segment.S0)
        intensities, phases = autd.link().drives_at(1, Segment.S0, 0)
        assert np.all(intensities == 0)

        with pytest.raises(InvalidDatagramTypeError):
            autd.group_send(lambda dev: dev.idx(), {0: 0})  # type: ignore[dict-item]


def test_group_check_only_for_enabled():
    autd: Controller[Audit]
    with create_controller() as autd:
        check = np.zeros(autd.num_devices(), dtype=bool)

        autd.geometry()[0].enable = False

        def key_map(dev: Device) -> int:
            check[dev.idx()] = True
            return 0

        autd.group_send(key_map, {0: (Sine(freq=150 * Hz, option=SineOption()), Uniform(intensity=EmitIntensity(0x80), phase=Phase(0x90)))})

        assert not check[0]
        assert check[1]

        mod = autd.link().modulation_buffer(0, Segment.S0)
        intensities, phases = autd.link().drives_at(0, Segment.S0, 0)
        assert np.all(intensities == 0)
        assert np.all(phases == 0)

        mod = autd.link().modulation_buffer(1, Segment.S0)
        assert len(mod) == 80
        intensities, phases = autd.link().drives_at(1, Segment.S0, 0)
        assert np.all(intensities == 0x80)
        assert np.all(phases == 0x90)


def test_clear():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send((Static(), Uniform(intensity=EmitIntensity(0xFF), phase=Phase(0x90))))

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0xFF)
            assert np.all(phases == 0x90)

        autd.send(Clear())

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)


def test_synchronize():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send(Synchronize())


def test_force_fan():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert not autd.link().is_force_fan(dev.idx())

        autd.send(ForceFan(lambda dev: dev.idx() == 0))
        assert autd.link().is_force_fan(0)
        assert not autd.link().is_force_fan(1)

        autd.send(ForceFan(lambda dev: dev.idx() == 1))
        assert not autd.link().is_force_fan(0)
        assert autd.link().is_force_fan(1)


def test_geometry():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.num_transducers() == 249 * 2
        assert autd[0].num_transducers() == 249


def test_tracing():
    tracing_init()
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.close()



from pyautd3.gain.holo.constraint import EmissionConstraint

from .amplitude import Amplitude, Pa, dB
from .backend_nalgebra import NalgebraBackend
from .greedy import Greedy, GreedyOption
from .gs import GS, GSOption
from .gspat import GSPAT, GSPATOption
from .lm import LM, LMOption
from .naive import Naive, NaiveOption

__all__ = [
    "GS",
    "GSPAT",
    "LM",
    "Amplitude",
    "EmissionConstraint",
    "GSOption",
    "GSPATOption",
    "Greedy",
    "GreedyOption",
    "LMOption",
    "Naive",
    "NaiveOption",
    "NalgebraBackend",
    "Pa",
    "dB",
]



from abc import ABCMeta, abstractmethod
from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, TransitionModeWrap

__all__ = []  # type: ignore[var-annotated]

DS = TypeVar("DS", bound="DatagramS")
P = TypeVar("P")


class DatagramS(Generic[P], metaclass=ABCMeta):
    @abstractmethod
    def _into_segment(self: Self, ptr: P, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        pass

    @abstractmethod
    def _raw_ptr(self: Self, geometry: Geometry) -> P:
        pass


class WithSegment(Datagram, Generic[DS]):
    inner: DS
    segment: Segment
    transitiom_mode: TransitionModeWrap | None

    def __init__(self: Self, inner: DS, segment: Segment, transition_mode: TransitionModeWrap | None) -> None:
        super().__init__()
        self.inner = inner
        self.segment = segment
        self.transitiom_mode = transition_mode

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return self.inner._into_segment(self.inner._raw_ptr(geometry), self.segment, self.transitiom_mode)



import numpy as np

from pyautd3.driver.defined.angle import Angle
from pyautd3.native_methods.autd3capi import NativeMethods as Base


class EulerAngles:
    def __new__(cls: type["EulerAngles"]) -> "EulerAngles":
        raise NotImplementedError

    @staticmethod
    def XYZ(x: Angle, y: Angle, z: Angle) -> np.ndarray:  # noqa: N802
        return Base().rotation_from_euler_xyz(x.radian(), y.radian(), z.radian()).ndarray()

    @staticmethod
    def ZYZ(z1: Angle, y: Angle, z2: Angle) -> np.ndarray:  # noqa: N802
        return Base().rotation_from_euler_zyz(z1.radian(), y.radian(), z2.radian()).ndarray()



import ctypes
from collections.abc import Iterable
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

M = TypeVar("M", bound=Modulation)


class Fir(Modulation, Generic[M]):
    target: M
    coef: np.ndarray

    def __init__(self: Self, target: M, coef: Iterable[float]) -> None:
        self.target = target
        self.coef = np.fromiter(coef, dtype=ctypes.c_float)

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_with_fir(
            self.target._modulation_ptr(),
            self.coef.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            len(self.coef),
        )



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

import pyautd3.native_methods.autd3capi_driver as consts


class AUTD3:
    pos: np.ndarray
    rot: np.ndarray

    def __init__(self: Self, pos: ArrayLike | None = None, rot: ArrayLike | None = None) -> None:
        self.pos = np.array(pos) if pos is not None else np.zeros(3)
        self.rot = np.array(rot) if rot is not None else np.array([1, 0, 0, 0])

    TRANS_SPACING: float = consts.TRANS_SPACING_MM
    DEVICE_WIDTH: float = consts.DEVICE_WIDTH_MM
    DEVICE_HEIGHT: float = consts.DEVICE_HEIGHT_MM
    NUM_TRANS_X: int = consts.NUM_TRANS_IN_X
    NUM_TRANS_Y: int = consts.NUM_TRANS_IN_Y
    NUM_TRANS_IN_UNIT: int = consts.NUM_TRANS_IN_UNIT



from abc import ABCMeta, abstractmethod
from typing import Self, TypeVar

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.with_loop_behavior import DatagramL
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, LoopBehavior, ModulationPtr, TransitionModeWrap

__all__ = []  # type: ignore[var-annotated]

M = TypeVar("M", bound="Modulation")


class Modulation(
    DatagramS[ModulationPtr],
    DatagramL[ModulationPtr],
    Datagram,
    metaclass=ABCMeta,
):
    def __init__(self: Self) -> None:
        super().__init__()

    def _raw_ptr(self: Self, _: Geometry) -> ModulationPtr:
        return self._modulation_ptr()

    def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
        return Base().modulation_into_datagram(self._modulation_ptr())

    def _into_segment(
        self: Self,
        ptr: ModulationPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
    ) -> DatagramPtr:
        return Base().modulation_into_datagram_with_segment(
            ptr,
            segment,
            transition_mode or TransitionMode.NONE,
        )

    def _into_loop_behavior(
        self: Self,
        ptr: ModulationPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return Base().modulation_into_datagram_with_loop_behavior(
            ptr,
            segment,
            transition_mode or TransitionMode.NONE,
            loop_behavior,
        )

    @abstractmethod
    def _modulation_ptr(self: Self) -> ModulationPtr:
        pass

    def sampling_config(self: Self) -> SamplingConfig:
        return SamplingConfig(Base().modulation_sampling_config(self._modulation_ptr()))



from pyautd3.driver.firmware.fpga import EmitIntensity
from pyautd3.native_methods.autd3capi_gain_holo import EmissionConstraintWrap
from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.utils import ConstantADT

__all__ = ["EmissionConstraint"]


class EmissionConstraint(metaclass=ConstantADT):
    def __new__(cls: type["EmissionConstraint"]) -> "EmissionConstraint":
        raise NotImplementedError

    Normalize: EmissionConstraintWrap = GainHolo().gain_holo_constraint_normalize()

    @staticmethod
    def Uniform(value: EmitIntensity) -> EmissionConstraintWrap:  # noqa: N802
        return GainHolo().gain_holo_constraint_uniform(value._inner())

    @staticmethod
    def Clamp(min_v: EmitIntensity, max_v: EmitIntensity) -> EmissionConstraintWrap:  # noqa: N802
        return GainHolo().gain_holo_constraint_clamp(min_v._inner(), max_v._inner())

    @staticmethod
    def Multiply(value: float) -> EmissionConstraintWrap:  # noqa: N802
        return GainHolo().gain_holo_constraint_multiply(value)



from abc import ABCMeta, abstractmethod
from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, LoopBehavior, TransitionModeWrap

__all__ = []  # type: ignore[var-annotated]

DL = TypeVar("DL", bound="DatagramL")
P = TypeVar("P")


class DatagramL(Generic[P], metaclass=ABCMeta):
    @abstractmethod
    def _into_loop_behavior(
        self: Self,
        ptr: P,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        pass

    @abstractmethod
    def _raw_ptr(self: Self, geometry: Geometry) -> P:
        pass


class WithLoopBehavior(Datagram, Generic[DL]):
    inner: DL
    segment: Segment
    transitiom_mode: TransitionModeWrap | None
    loop_behavior: LoopBehavior

    def __init__(self: Self, inner: DL, segment: Segment, transition_mode: TransitionModeWrap | None, loop_behavior: LoopBehavior) -> None:
        super().__init__()
        self.inner = inner
        self.segment = segment
        self.transitiom_mode = transition_mode
        self.loop_behavior = loop_behavior

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return self.inner._into_loop_behavior(self.inner._raw_ptr(geometry), self.segment, self.transitiom_mode, self.loop_behavior)



from collections.abc import Iterable
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.driver.datagram.gain import Gain
from pyautd3.gain.holo.backend import Backend

from .amplitude import Amplitude

H = TypeVar("H", bound="Holo")


class Holo(Gain, Generic[H]):
    foci: list[tuple[np.ndarray, Amplitude]]

    def __init__(self: Self, foci: Iterable[tuple[np.ndarray, Amplitude]]) -> None:
        self.foci = list(foci)


class HoloWithBackend(Holo[H], Generic[H]):
    backend: Backend

    def __init__(
        self: Self,
        foci: Iterable[tuple[np.ndarray, Amplitude]],
        backend: Backend,
    ) -> None:
        super().__init__(foci)
        self.backend = backend



from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import ModulationCachePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

M = TypeVar("M", bound=Modulation)


class Cache(Modulation, Generic[M]):
    _m: M
    _ptr: ModulationCachePtr | None

    def __init__(self: Self, m: M) -> None:
        super().__init__()
        self._m = m
        self._ptr = None

    def _modulation_ptr(self: Self) -> ModulationPtr:
        if self._ptr is None:
            self._ptr = Base().modulation_cache(self._m._modulation_ptr())
        return Base().modulation_cache_clone(self._ptr)

    def __del__(self: Self) -> None:
        if self._ptr is not None:
            Base().modulation_cache_free(self._ptr)



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import FocusOption as FocusOption_
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Point3


class FocusOption:
    intensity: EmitIntensity
    phase_offset: Phase

    def __init__(self: Self, *, intensity: EmitIntensity | None = None, phase_offset: Phase | None = None) -> None:
        self.intensity = intensity or EmitIntensity.MAX
        self.phase_offset = phase_offset or Phase(0)

    def _inner(self: Self) -> FocusOption_:
        return FocusOption_(self.intensity._inner(), self.phase_offset._inner())


class Focus(Gain):
    pos: np.ndarray
    option: FocusOption

    def __init__(self: Self, pos: ArrayLike, option: FocusOption) -> None:
        super().__init__()
        self.pos = np.array(pos)
        self.option = option

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_focus(Point3(self.pos), self.option._inner())



from collections.abc import Iterator
from typing import Self

import numpy as np

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GeometryPtr

from .device import Device


class Geometry:
    _geometry_ptr: GeometryPtr
    _devices: list[Device]

    def __init__(self: Self, ptr: GeometryPtr) -> None:
        self._geometry_ptr = ptr
        self._devices = [Device(i, ptr) for i in range(int(Base().geometry_num_devices(self._geometry_ptr)))]

    def center(self: Self) -> np.ndarray:
        return Base().geometr_center(self._geometry_ptr).ndarray()

    def num_devices(self: Self) -> int:
        return int(Base().geometry_num_devices(self._geometry_ptr))

    def num_transducers(self: Self) -> int:
        return int(Base().geometry_num_transducers(self._geometry_ptr))

    def __getitem__(self: Self, key: int) -> Device:
        return self._devices[key]

    def __iter__(self: Self) -> Iterator[Device]:
        return iter(self._devices)

    def devices(self: Self) -> Iterator[Device]:
        return filter(lambda x: x.enable, self._devices)

    def set_sound_speed_from_temp(
        self: Self,
        temp: float,
        k: float = 1.4,
        r: float = 8.31446261815324,
        m: float = 28.9647e-3,
    ) -> None:
        for d in self.devices():
            d.set_sound_speed_from_temp(temp, k, r, m)

    def set_sound_speed(
        self: Self,
        c: float,
    ) -> None:
        for d in self.devices():
            d.sound_speed = c



import contextlib
from collections.abc import Callable

from .controller import Controller, SenderOption
from .controller.sleeper import SpinSleeper
from .driver.autd3_device import AUTD3
from .driver.datagram import (
    Clear,
    DebugSettings,
    DebugType,
    FixedCompletionTime,
    FixedUpdateRate,
    FociSTM,
    ForceFan,
    GainSTM,
    GainSTMOption,
    PhaseCorrection,
    PulseWidthEncoder,
    ReadsFPGAState,
    Silencer,
    SwapSegment,
    WithLoopBehavior,
    WithSegment,
)
from .driver.datagram.stm import (
    ControlPoint,
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
)
from .driver.defined import Hz, deg, kHz, rad
from .driver.firmware.fpga import Drive, EmitIntensity, LoopBehavior, Phase, SamplingConfig, TransitionMode
from .driver.geometry import Device, EulerAngles, Geometry, Transducer
from .ethercat import DcSysTime
from .gain import Bessel, BesselOption, Focus, FocusOption, Group, Null, Plane, PlaneOption, Uniform
from .link.nop import Nop
from .modulation import Sine, SineOption, Square, SquareOption, Static
from .native_methods.autd3 import GainSTMMode, GPIOIn, GPIOOut, ParallelMode, Segment, SilencerTarget
from .native_methods.autd3capi import NativeMethods as Base
from .native_methods.autd3capi_link_simulator import NativeMethods as Simulator
from .native_methods.autd3capi_link_twincat import NativeMethods as TwinCAT
from .native_methods.autd3capi_modulation_audio_file import NativeMethods as AudioFile
from .utils import Duration

_ext_tracing_init: list[Callable[[], None]] = []


def tracing_init() -> None:
    Base().tracing_init()
    with contextlib.suppress(BaseException):
        Simulator().link_simulator_tracing_init()
    with contextlib.suppress(BaseException):
        TwinCAT().link_twin_cat_tracing_init()
    with contextlib.suppress(BaseException):
        AudioFile().modulation_audio_file_tracing_init()
    for func in _ext_tracing_init:  # pragma: no cover
        func()  # pragma: no cover


__all__ = [
    "AUTD3",
    "Bessel",
    "BesselOption",
    "Clear",
    "ControlPoint",
    "ControlPoints1",
    "ControlPoints2",
    "ControlPoints3",
    "ControlPoints4",
    "ControlPoints5",
    "ControlPoints6",
    "ControlPoints7",
    "ControlPoints8",
    "Controller",
    "Custom",
    "DcSysTime",
    "DebugSettings",
    "DebugType",
    "Device",
    "Drive",
    "Duration",
    "EmitIntensity",
    "EulerAngles",
    "FixedCompletionTime",
    "FixedUpdateRate",
    "FociSTM",
    "Focus",
    "FocusOption",
    "ForceFan",
    "GPIOIn",
    "GPIOOut",
    "GainSTM",
    "GainSTMMode",
    "GainSTMOption",
    "Geometry",
    "Group",
    "Hz",
    "LoopBehavior",
    "Nop",
    "Null",
    "ParallelMode",
    "Phase",
    "PhaseCorrection",
    "Plane",
    "PlaneOption",
    "PulseWidthEncoder",
    "ReadsFPGAState",
    "SamplingConfig",
    "SamplingMode",
    "Segment",
    "SenderOption",
    "Silencer",
    "SilencerTarget",
    "Sine",
    "SineOption",
    "SpinSleeper",
    "Square",
    "SquareOption",
    "Static",
    "SwapSegment",
    "Transducer",
    "TransitionMode",
    "Uniform",
    "WithLoopBehavior",
    "WithSegment",
    "deg",
    "kHz",
    "phase_rad",
    "rad",
    "tracing_init",
]

__version__ = "30.0.1.1"
